// SharedMemoryWriter.cpp
// 2016.05 Kyoung Shin Park 
//
#include "QUANTA/QUANTAinit.hxx"
#include "QUANTA/QUANTAnet_datapack_c.hxx"
#include "SharedMemory.h"
#include "Semaphore.h"

#include <windows.h>
#include <iostream>
#include <string.h>

#define BUFFER_SIZE 4096
#define MAX_COUNT 30
static int gCount = 0;

// shared memory 
CSharedMemory shmem("_SHMEM_TEST");
char * shmbuf;
HANDLE segment_id = (HANDLE)3782;
// semaphore (mutex)
KSemaphore sem("_SEM_TEST");
// datapack
bool datapackActive = false;
static QUANTAnet_datapack_c dataPacker;

void parseCommandLine(int argc, char ** argv)
{
    for (int i = 1; i < argc; i++) {
		if (!strncmp(argv[i], "--", 2)) {
			if (!strncmp(argv[i], "--datapack", 10)) {
				datapackActive = true;
				printf("datapack active\n");
			}
		}
	}
}

// pack data and write data to shared memory
int packDataTo(char* buffer)
{
	int myInt = 20000000;
	float myFloat = 3.33f;
	float floatArray[4] = { 1.1f, 2.2f, 3.3f, 4.4f };
	double myDouble = 4.44444;
	long myLong = 1138;
	unsigned long myUnsignedLong = 4000000000;
	char myChar = 'A';
	char myStr[8];
	strcpy(myStr, "thx1138");
	char* fullString = "jar jar";

	int dataSize = dataPacker.sizeof_int()	// int
		+ dataPacker.sizeof_float()			// float
		+ dataPacker.sizeof_float(4)		// float array
		+ dataPacker.sizeof_double()		// double
		+ dataPacker.sizeof_long()			// long
		+ dataPacker.sizeof_unsignedLong()	// unsigned long
		+ dataPacker.sizeof_char()			// char
		+ dataPacker.sizeof_char(8)			// char array
		+ dataPacker.sizeof_string(32);		// string

	sem.lock(); // shared memory lock
	printf("Send Int: %d\n", gCount);
	printf("Send Float: %f\n", myFloat);
	printf("Send Float Array - with the floats: ");
	for (int i = 0; i<4; i++) printf("%f\t", floatArray[i]);
	printf("\nSend Double: %lf\n", myDouble);
	printf("Send Long: %ld\n", myLong);
	printf("Send Unsigned Long: %lu\n", myUnsignedLong);
	printf("Send Char: %c\n", myChar);
	printf("Send Char[]: %s\n", myStr);
	printf("Send String: %s\n", fullString);
	dataPacker.initPack(buffer, dataSize);
	dataPacker.packInt(gCount);
	dataPacker.packFloat(myFloat);
	dataPacker.packFloatArray(floatArray, 4);
	dataPacker.packDouble(myDouble);
	dataPacker.packLong(myLong);
	dataPacker.packUnsignedLong(myUnsignedLong);
	dataPacker.packChar(myChar);
	dataPacker.pack(myStr, 8);
	dataPacker.packString(fullString);
	gCount++;
	sem.unlock(); // shared memory unlock

	int realSize = dataPacker.getBufferFilledSize();
	printf("\n\nSIZE: %d/%d\n", realSize, dataSize);

	return realSize;
}

int writeDataTo(char* buffer)
{
	sem.lock(); // shared memory lock
	printf("%d %d\n", gCount, gCount);
	sprintf(buffer, "%d %d", gCount, gCount);
	gCount++;
	sem.unlock(); // shared memory unlock
	return sizeof(int);
}

int main(int argc, char ** argv)
{
	parseCommandLine(argc, argv);

	// create semaphore
	sem.create();
	// initialize shared memory unit & allocate shared memory segment
	segment_id = shmem.getShm((key_t)segment_id, BUFFER_SIZE, S_IRUSR | S_IWUSR | IPC_CREAT);
	printf("shmem id=%d\n", (int)segment_id);

	// attach the shared memory segment
	shmbuf = (char *) shmem.attachShm(segment_id, NULL, 0);
	printf("shared memory attached\n");

	// while loop
	while (gCount < MAX_COUNT) {
		int dataSize = 0;
		if (datapackActive) {
			dataSize = packDataTo(shmbuf);
		}
		else {
			dataSize = writeDataTo(shmbuf);
		}
		printf("Data Shared Memory ߽ϴ: DataSize=%d\n", dataSize);
		Sleep(500);
	}

	// detach the shared memory segment
	shmem.detachShm(shmbuf);
	// remove the shared memory segment
	shmem.cleanShm(segment_id, IPC_RMID, NULL);
	sem.close();
	printf("shared memory detached & cleanup \n");

	return 0;
}

